==> intrepyd-0.10.0.tar.gz: intrepyd-0.10.0/intrepyd/parser.py <==
"""
Copyright (C) 2021 Roberto Bruttomesso <roberto.bruttomesso@gmail.com>

This file is distributed under the terms of the 3-clause BSD License.
A copy of the license can be found in the root directory or at
https://opensource.org/licenses/BSD-3-Clause.

Author: Roberto Bruttomesso <roberto.bruttomesso@gmail.com>
  Date: 27/03/2021
"""

from intrepyd.context import Context

class ParseError(Exception):
    def __init__(self, message):
        super().__init__(message)
        self._message = message

    def message(self):
        return self._message

class Parser(object):
    def __init__(self):
        self._ctx = Context()
        self._special_ops = set(['set_latch_init_next'])
        self._initialize_parsing_functions()
        self._initialize_types()
        self._ctx.nets['true'] = self._ctx.mk_true()
        self._ctx.nets['false'] = self._ctx.mk_false()

    def _initialize_types(self):
        self._name_to_type = {
          'bool': self._ctx.mk_boolean_type(),
          'int8': self._ctx.mk_int8_type(),
          'int16': self._ctx.mk_int16_type(),
          'int32': self._ctx.mk_int32_type(),
          'uint8': self._ctx.mk_uint8_type(),
          'uint16': self._ctx.mk_uint16_type(),
          'uint32': self._ctx.mk_uint32_type(),
          'float16': self._ctx.mk_float16_type(),
          'float32': self._ctx.mk_float32_type(),
          'float64': self._ctx.mk_float64_type()
        }

    def _initialize_parsing_functions(self):
        self._op_to_unary_func = {}
        self._op_to_binary_func = {}
        self._op_to_ternary_func = {}
        self._op_to_unary_func['not'] = self._ctx.mk_not
        self._op_to_unary_func['minus'] = self._ctx.mk_minus
        self._op_to_unary_func['to_int8'] = self._ctx.mk_cast_to_int8
        self._op_to_unary_func['to_int16'] = self._ctx.mk_cast_to_int16
        self._op_to_unary_func['to_int32'] = self._ctx.mk_cast_to_int32
        self._op_to_unary_func['to_uint8'] = self._ctx.mk_cast_to_uint8
        self._op_to_unary_func['to_uint16'] = self._ctx.mk_cast_to_uint16
        self._op_to_unary_func['to_uint32'] = self._ctx.mk_cast_to_uint32
        self._op_to_binary_func['and'] = self._ctx.mk_and
        self._op_to_binary_func['or'] = self._ctx.mk_or
        self._op_to_binary_func['xor'] = self._ctx.mk_xor
        self._op_to_binary_func['iff'] = self._ctx.mk_iff
        self._op_to_binary_func['implies'] = self._ctx.mk_implies
        self._op_to_binary_func['eq'] = self._ctx.mk_eq
        self._op_to_binary_func['neq'] = self._ctx.mk_neq
        self._op_to_binary_func['leq'] = self._ctx.mk_leq
        self._op_to_binary_func['geq'] = self._ctx.mk_geq
        self._op_to_binary_func['lt'] = self._ctx.mk_lt
        self._op_to_binary_func['gt'] = self._ctx.mk_gt
        self._op_to_binary_func['add'] = self._ctx.mk_add
        self._op_to_binary_func['mul'] = self._ctx.mk_mul
        self._op_to_binary_func['div'] = self._ctx.mk_div
        self._op_to_binary_func['sub'] = self._ctx.mk_sub
        self._op_to_binary_func['mod'] = self._ctx.mk_mod
        self._op_to_ternary_func['ite'] = self._ctx.mk_ite

    def parse_file(self, filepath):
        with open(filepath, "rt") as f:
            return self.parse_stream(f)

    def parse_stream(self, f):
        return self._parse(f)

    def _parse(self, file):
        lineno = 1
        line = file.readline()
        while line:
            self._parse_line(line, lineno)
            lineno += 1
            line = file.readline()
        return self._ctx

    def _parse_line(self, line, lineno):
        line = line.strip()
        if len(line) == 0:
            return
        if line[0] == '#':
            return
        tokens = line.split()
        if len(tokens) < 4:
            raise ParseError('Parse error at line {}: less than four tokens'.format(lineno))
        if len(tokens) > 6:
            raise ParseError('Parse error at line {}: more than six tokens'.format(lineno))
        if tokens[0] in self._special_ops:
            self._parse_special_op(line, lineno)
            return
        identifier = tokens[0]
        if identifier in self._ctx.nets:
            raise ParseError('Parse error at line {}: identifier {} already used'.format(lineno, identifier))
        if tokens[1] != '=':
            raise ParseError('Parse error at line {}: expected =, found {}'.format(lineno, tokens[1]))
        operator = tokens[2]
        x = None
        y = None
        z = None
        net = None
        if len(tokens) >= 4:
            x = tokens[3]
        if operator == 'latch' or operator == 'input':
            if len(tokens) != 4:
                raise ParseError('Parse error at line {}: unexpected token {}'.format(lineno, tokens[4]))
            if x not in self._name_to_type:
                raise ParseError('Parse error at line {}: unrecognized type {}'.format(lineno, x))
            x = self._name_to_type[x]
            if operator == 'latch':
                net = self._ctx.mk_latch(identifier, x)
            if operator == 'input':
                net = self._ctx.mk_input(identifier, x)
            assert(net is not None)
            assert(self._ctx.nets[identifier] == net)
            return
        if len(tokens) >= 5:
            y = tokens[4]
        if operator == 'number':
            if len(tokens) != 5:
                raise ParseError('Parse error at line {}: unexpected token {}'.format(lineno, tokens[5]))
            if y not in self._name_to_type:
                raise ParseError('Parse error at line {}: unrecognized type {}'.format(lineno, x))
            y = self._name_to_type[y]
            net = self._ctx.mk_number(x, y, identifier)
            assert(net is not None)
            assert(self._ctx.nets[identifier] == net)
            return
        if len(tokens) >= 4:
            if x not in self._ctx.nets:
                raise ParseError('Parse error at line {}: identifier {} not found'.format(lineno, x))
            x = self._ctx.nets[x]
        if len(tokens) >= 5:
            if y not in self._ctx.nets:
                raise ParseError('Parse error at line {}: identifier {} not found'.format(lineno, y))
            y = self._ctx.nets[y]
        if len(tokens) >= 6:
            z = tokens[5]
            if z not in self._ctx.nets:
                raise ParseError('Parse error at line {}: identifier {} not found'.format(lineno, z))
            z = self._ctx.nets[z]
        # Unary operators
        if len(tokens) == 4:
            assert(x is not None)
            if operator not in self._op_to_unary_func:
                raise ParseError('Parse error at line {}: operator {} not found'.format(lineno, operator))
            net = self._op_to_unary_func[operator](x, identifier)
        # Binary operators
        if len(tokens) == 5:
            assert(x is not None)
            assert(y is not None)
            if operator not in self._op_to_binary_func:
                raise ParseError('Parse error at line {}: operator {} not found'.format(lineno, operator))
            net = self._op_to_binary_func[operator](x, y, identifier)
        # Ternary operators
        if len(tokens) == 6:
            assert(x is not None)
            assert(y is not None)
            assert(z is not None)
            if operator not in self._op_to_ternary_func:
                raise ParseError('Parse error at line {}: operator {} not found'.format(lineno, operator))
            net = self._op_to_ternary_func[operator](x, y, z, identifier)
        assert(net is not None)
        assert(self._ctx.nets[identifier] == net)

    def _parse_special_op(self, line, lineno):
        tokens = line.split()
        operator = tokens[0]
        if operator == 'set_latch_init_next':
            l = tokens[1]
            i = tokens[2]
            n = tokens[3]
            if l not in self._ctx.nets:
                raise ParseError('Parse error at line {}: identifier {} not found'.format(lineno, l))
            if i not in self._ctx.nets:
                raise ParseError('Parse error at line {}: identifier {} not found'.format(lineno, i))
            if n not in self._ctx.nets:
                raise ParseError('Parse error at line {}: identifier {} not found'.format(lineno, n))
            l = self._ctx.nets[l]
            i = self._ctx.nets[i]
            n = self._ctx.nets[n]
            self._ctx.set_latch_init_next(l, i, n)

==> intrepyd-0.10.3.tar.gz: intrepyd-0.10.3/intrepyd/parser.py <==
"""
A parser for the Intrepid's language
"""

from intrepyd.context import Context

class ParseError(Exception):
    """
    Encapsulates a parsing error
    """
    def __init__(self, message):
        super().__init__(message)
        self._message = message

    def message(self):
        """
        Retrieves the error message
        """
        return self._message

class Parser:
    """
    A parser for Intrepid's syntax
    """
    def __init__(self):
        self._ctx = Context()
        self._special_ops = set(['set_latch_init_next'])
        self._initialize_parsing_functions()
        self._initialize_types()
        self._ctx.nets['true'] = self._ctx.mk_true()
        self._ctx.nets['false'] = self._ctx.mk_false()

    def _initialize_types(self):
        self._name_to_type = {
          'bool': self._ctx.mk_boolean_type(),
          'int8': self._ctx.mk_int8_type(),
          'int16': self._ctx.mk_int16_type(),
          'int32': self._ctx.mk_int32_type(),
          'int64': self._ctx.mk_int64_type(),
          'uint8': self._ctx.mk_uint8_type(),
          'uint16': self._ctx.mk_uint16_type(),
          'uint32': self._ctx.mk_uint32_type(),
          'uint64': self._ctx.mk_uint64_type(),
          'float16': self._ctx.mk_float16_type(),
          'float32': self._ctx.mk_float32_type(),
          'float64': self._ctx.mk_float64_type()
        }

    def _initialize_parsing_functions(self):
        self._op_to_unary_func = {}
        self._op_to_binary_func = {}
        self._op_to_ternary_func = {}
        self._op_to_unary_func['not'] = self._ctx.mk_not
        self._op_to_unary_func['minus'] = self._ctx.mk_minus
        self._op_to_unary_func['to_int8'] = self._ctx.mk_cast_to_int8
        self._op_to_unary_func['to_int16'] = self._ctx.mk_cast_to_int16
        self._op_to_unary_func['to_int32'] = self._ctx.mk_cast_to_int32
        self._op_to_unary_func['to_uint8'] = self._ctx.mk_cast_to_uint8
        self._op_to_unary_func['to_uint16'] = self._ctx.mk_cast_to_uint16
        self._op_to_unary_func['to_uint32'] = self._ctx.mk_cast_to_uint32
        self._op_to_binary_func['and'] = self._ctx.mk_and
        self._op_to_binary_func['or'] = self._ctx.mk_or
        self._op_to_binary_func['xor'] = self._ctx.mk_xor
        self._op_to_binary_func['iff'] = self._ctx.mk_iff
        self._op_to_binary_func['implies'] = self._ctx.mk_implies
        self._op_to_binary_func['eq'] = self._ctx.mk_eq
        self._op_to_binary_func['neq'] = self._ctx.mk_neq
        self._op_to_binary_func['leq'] = self._ctx.mk_leq
        self._op_to_binary_func['geq'] = self._ctx.mk_geq
        self._op_to_binary_func['lt'] = self._ctx.mk_lt
        self._op_to_binary_func['gt'] = self._ctx.mk_gt
        self._op_to_binary_func['add'] = self._ctx.mk_add
        self._op_to_binary_func['mul'] = self._ctx.mk_mul
        self._op_to_binary_func['div'] = self._ctx.mk_div
        self._op_to_binary_func['sub'] = self._ctx.mk_sub
        self._op_to_binary_func['mod'] = self._ctx.mk_mod
        self._op_to_ternary_func['ite'] = self._ctx.mk_ite

    def parse_file(self, filepath):
        """
        Parses the file, and returns the context
        """
        with open(filepath, "rt") as file:
            return self.parse_stream(file)

    def parse_stream(self, stream):
        """
        Parses the stream, and returns the context
        """
        return self._parse(stream)

    def _parse(self, file):
        lineno = 1
        line = file.readline()
        while line:
            self._parse_line(line, lineno)
            lineno += 1
            line = file.readline()
        return self._ctx

    def _parse_line(self, line, lineno):
        line = line.strip()
        if len(line) == 0:
            return
        if line[0] == '#':
            return
        tokens = line.split()
        if len(tokens) < 4:
            raise ParseError('Parse error at line {}: less than four tokens'.format(lineno))
        if len(tokens) > 6:
            raise ParseError('Parse error at line {}: more than six tokens'.format(lineno))
        if tokens[0] in self._special_ops:
            self._parse_special_op(line, lineno)
            return
        identifier = tokens[0]
        if identifier in self._ctx.nets:
            raise ParseError('Parse error at line {}: identifier {} already used'\
                             .format(lineno, identifier))
        if tokens[1] != '=':
            raise ParseError('Parse error at line {}: expected =, found {}'\
                             .format(lineno, tokens[1]))
        operator = tokens[2]
        x = None
        y = None
        z = None
        net = None
        if len(tokens) >= 4:
            x = tokens[3]
        if operator in ('latch', 'input'):
            if len(tokens) != 4:
                raise ParseError('Parse error at line {}: unexpected token {}'\
                                 .format(lineno, tokens[4]))
            if x not in self._name_to_type:
                raise ParseError('Parse error at line {}: unrecognized type {}'\
                                 .format(lineno, x))
            x = self._name_to_type[x]
            if operator == 'latch':
                net = self._ctx.mk_latch(identifier, x)
            if operator == 'input':
                net = self._ctx.mk_input(identifier, x)
            assert net is not None
            assert self._ctx.nets[identifier] == net
            return
        if len(tokens) >= 5:
            y = tokens[4]
        if operator == 'number':
            if len(tokens) != 5:
                raise ParseError('Parse error at line {}: unexpected token {}'\
                                 .format(lineno, tokens[5]))
            if y not in self._name_to_type:
                raise ParseError('Parse error at line {}: unrecognized type {}'.format(lineno, x))
            y = self._name_to_type[y]
            net = self._ctx.mk_number(x, y, identifier)
            assert net is not None
            assert self._ctx.nets[identifier] == net
            return
        if len(tokens) >= 4:
            if x not in self._ctx.nets:
                raise ParseError('Parse error at line {}: identifier {} not found'\
                                 .format(lineno, x))
            x = self._ctx.nets[x]
        if len(tokens) >= 5:
            if y not in self._ctx.nets:
                raise ParseError('Parse error at line {}: identifier {} not found'\
                                 .format(lineno, y))
            y = self._ctx.nets[y]
        if len(tokens) >= 6:
            z = tokens[5]
            if z not in self._ctx.nets:
                raise ParseError('Parse error at line {}: identifier {} not found'\
                                 .format(lineno, z))
            z = self._ctx.nets[z]
        # Unary operators
        if len(tokens) == 4:
            assert x is not None
            if operator not in self._op_to_unary_func:
                raise ParseError('Parse error at line {}: operator {} not found'\
                                 .format(lineno, operator))
            net = self._op_to_unary_func[operator](x, identifier)
        # Binary operators
        if len(tokens) == 5:
            assert x is not None
            assert y is not None
            if operator not in self._op_to_binary_func:
                raise ParseError('Parse error at line {}: operator {} not found'\
                                 .format(lineno, operator))
            net = self._op_to_binary_func[operator](x, y, identifier)
        # Ternary operators
        if len(tokens) == 6:
            assert x is not None
            assert y is not None
            assert z is not None
            if operator not in self._op_to_ternary_func:
                raise ParseError('Parse error at line {}: operator {} not found'\
                                 .format(lineno, operator))
            net = self._op_to_ternary_func[operator](x, y, z, identifier)
        assert net is not None
        assert self._ctx.nets[identifier] == net

    def _parse_special_op(self, line, lineno):
        tokens = line.split()
        operator = tokens[0]
        if operator == 'set_latch_init_next':
            lat = tokens[1]
            ini = tokens[2]
            nex = tokens[3]
            if lat not in self._ctx.nets:
                raise ParseError('Parse error at line {}: identifier {} not found'\
                                 .format(lineno, lat))
            if ini not in self._ctx.nets:
                raise ParseError('Parse error at line {}: identifier {} not found'\
                                 .format(lineno, ini))
            if nex not in self._ctx.nets:
                raise ParseError('Parse error at line {}: identifier {} not found'\
                                 .format(lineno, nex))
            lat = self._ctx.nets[lat]
            ini = self._ctx.nets[ini]
            nex = self._ctx.nets[nex]
            self._ctx.set_latch_init_next(lat, ini, nex)
